'''
A collection of utilities that aren't large enough to have their own modules.

Each file in this directory is a useful standalone tool, that may or may not
someday be moved into its own separate package.
'''